#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2013-2023 by Björn Johansson.  All rights reserved.
# This code is part of the Python-dna distribution and governed by its
# license.  Please see the LICENSE.txt file that should have been included
# as part of this package.
"""This module contain functions for primer design for various purposes.

- :func:primer_design for designing primers for a sequence or a matching primer for an existing primer. Returns an :class:`Amplicon` object (same as the :mod:`amplify` module returns).

- :func:assembly_fragments Adds tails to primers for a linear assembly through homologous recombination or Gibson assembly.

- :func:circular_assembly_fragments Adds tails to primers for a circular assembly through homologous recombination or Gibson assembly.

"""

from pydna.tm import tm_default as _tm_default
import math as _math

# import os as _os
import copy as _copy
from pydna.amplicon import Amplicon as _Amplicon
from pydna.amplify import Anneal as _Anneal
from pydna.amplify import pcr as _pcr
from pydna.dseqrecord import Dseqrecord as _Dseqrecord
from pydna.primer import Primer as _Primer

# import logging as _logging
import operator as _operator
from typing import Tuple

# _module_logger = _logging.getLogger("pydna." + __name__)


def _design_primer(
    target_tm: float,
    template: _Dseqrecord,
    limit: int,
    tm_func,
    starting_length: int = 0,
) -> Tuple[float, str]:
    """returns a tuple (temp, primer)"""

    if starting_length < limit:
        starting_length = limit

    length = starting_length
    tlen = len(template)

    tmp = max(0, tm_func(str(template.seq[:length])))

    p = str(template.seq[:length])

    if tmp < target_tm:
        condition = _operator.le
        increment = 1
    else:
        condition = _operator.ge
        increment = -1
    while condition(tmp, target_tm):
        prev_temp = tmp
        prev_primer = p
        length += increment
        p = str(template.seq[:length])
        tmp = tm_func(p)
        if length >= tlen:
            break
        # Never go below the limit
        if length < limit:
            return template.seq[:limit]

    # _module_logger.debug(((p, tmp), (prev_primer, prev_temp)))
    if abs(target_tm - tmp) < abs(target_tm - prev_temp):
        return p
    else:
        return prev_primer


def primer_design(
    template,
    fp=None,
    rp=None,
    limit=13,
    target_tm=55.0,
    tm_func=_tm_default,
    estimate_function=None,
    **kwargs,
):
    """This function designs a forward primer and a reverse primer for PCR amplification
    of a given template sequence.

    The template argument is a Dseqrecord object or equivalent containing the template sequence.

    The optional fp and rp arguments can contain an existing primer for the sequence (either the forward or reverse primer).
    One or the other primers can be specified, not both (since then there is nothing to design!, use the pydna.amplify.pcr function instead).

    The limit argument is the minimum length of the primer. The default value is 13.

    If one of the primers is given, the other primer is designed to match in terms of Tm.
    If both primers are designed, they will be designed to target_tm

    tm_func is a function that takes an ascii string representing an oligonuceotide as argument and returns a float.
    Some useful functions can be found in the :mod:`pydna.tm` module, but can be substituted for a custom made function.

    estimate_function is a tm_func-like function that is used to get a first guess for the primer design, that is then used as starting
    point for the final result. This is useful when the tm_func function is slow to calculate (e.g. it relies on an
    external API, such as the NEB primer design API). The estimate_function should be faster than the tm_func function.
    The default value is `None`.
    To use the default `tm_func` as estimate function to get the NEB Tm faster, you can do:
    `primer_design(dseqr, target_tm=55, tm_func=tm_neb, estimate_function=tm_default)`.

    The function returns a pydna.amplicon.Amplicon class instance. This object has
    the object.forward_primer and object.reverse_primer properties which contain the designed primers.


    Parameters
    ----------

    template : pydna.dseqrecord.Dseqrecord
        a Dseqrecord object. The only required argument.

    fp, rp : pydna.primer.Primer, optional
        optional pydna.primer.Primer objects containing one primer each.

    target_tm : float, optional
        target tm for the primers, set to 55°C by default.

    tm_func : function
        Function used for tm calculation. This function takes an ascii string
        representing an oligonuceotide as argument and returns a float.
        Some useful functions can be found in the :mod:`pydna.tm` module, but can be
        substituted for a custom made function.

    Returns
    -------
    result : Amplicon

    Examples
    --------

    >>> from pydna.dseqrecord import Dseqrecord
    >>> t=Dseqrecord("atgactgctaacccttccttggtgttgaacaagatcgacgacatttcgttcgaaacttacgatg")
    >>> t
    Dseqrecord(-64)
    >>> from pydna.design import primer_design
    >>> ampl = primer_design(t)
    >>> ampl
    Amplicon(64)
    >>> ampl.forward_primer
    f64 17-mer:5'-atgactgctaacccttc-3'
    >>> ampl.reverse_primer
    r64 18-mer:5'-catcgtaagtttcgaacg-3'
    >>> print(ampl.figure())
    5atgactgctaacccttc...cgttcgaaacttacgatg3
                         ||||||||||||||||||
                        3gcaagctttgaatgctac5
    5atgactgctaacccttc3
     |||||||||||||||||
    3tactgacgattgggaag...gcaagctttgaatgctac5
    >>> pf = "GGATCC" + ampl.forward_primer
    >>> pr = "GGATCC" + ampl.reverse_primer
    >>> pf
    f64 23-mer:5'-GGATCCatgactgct..ttc-3'
    >>> pr
    r64 24-mer:5'-GGATCCcatcgtaag..acg-3'
    >>> from pydna.amplify import pcr
    >>> pcr_prod = pcr(pf, pr, t)
    >>> print(pcr_prod.figure())
          5atgactgctaacccttc...cgttcgaaacttacgatg3
                               ||||||||||||||||||
                              3gcaagctttgaatgctacCCTAGG5
    5GGATCCatgactgctaacccttc3
           |||||||||||||||||
          3tactgacgattgggaag...gcaagctttgaatgctac5
    >>> print(pcr_prod.seq)
    GGATCCatgactgctaacccttccttggtgttgaacaagatcgacgacatttcgttcgaaacttacgatgGGATCC
    >>> from pydna.primer import Primer
    >>> pf = Primer("atgactgctaacccttccttggtgttg", id="myprimer")
    >>> ampl = primer_design(t, fp = pf)
    >>> ampl.forward_primer
    myprimer 27-mer:5'-atgactgctaaccct..ttg-3'
    >>> ampl.reverse_primer
    r64 32-mer:5'-catcgtaagtttcga..atc-3'
    """

    def design(target_tm, template):
        if estimate_function:
            first_guess = _design_primer(target_tm, template, limit, estimate_function)
            return _design_primer(target_tm, template, limit, tm_func, len(first_guess))
        else:
            return _design_primer(target_tm, template, limit, tm_func)

    if not fp and not rp:
        # _module_logger.debug("no primer given, design forward primer:")
        fp = _Primer((design(target_tm, template)))
        target_tm = tm_func(str(fp.seq))
        # _module_logger.debug("no primer given, design reverse primer:")
        rp = _Primer(design(target_tm, template.reverse_complement()))
    elif fp and not rp:
        try:
            fp = _Anneal((fp,), template).forward_primers.pop()
        except IndexError:
            raise ValueError("Forward primer does not anneal")
        except Exception:  # pragma: no cover
            print("Unexpected error")
            raise
        target_tm = tm_func(fp.footprint)
        # _module_logger.debug("forward primer given, design reverse primer:")
        rp = _Primer(design(target_tm, template.reverse_complement()))
    elif not fp and rp:
        try:
            rp = _Anneal((rp,), template).reverse_primers.pop()
        except IndexError:
            raise ValueError("Reverse primer does not anneal")
        except Exception:  # pragma: no cover
            print("Unexpected error")
            raise
        target_tm = tm_func(rp.footprint)
        # _module_logger.debug("reverse primer given, design forward primer:")
        fp = _Primer(design(target_tm, template))
    else:
        raise ValueError("Specify maximum one of the two primers.")

    if fp.id == "id":  # <unknown id>
        fp.id = "f{}".format(len(template))

    if rp.id == "id":
        rp.id = "r{}".format(len(template))

    if fp.name == "name":
        fp.name = "f{}".format(len(template))

    if rp.name == "name":
        rp.name = "r{}".format(len(template))

    fp.description = fp.id + " " + template.accession
    rp.description = rp.id + " " + template.accession

    ampl = _Anneal((fp, rp), template, limit=limit)

    prod = ampl.products[0] if ampl.products else _Amplicon("")

    if len(ampl.products) > 1:
        import warnings as _warnings
        from pydna import _PydnaWarning

        _warnings.warn(
            "designed primers do not yield a unique PCR product", _PydnaWarning
        )

    return prod


def assembly_fragments(f, overlap=35, maxlink=40, circular=False):
    """This function return a list of :mod:`pydna.amplicon.Amplicon` objects where
    primers have been modified with tails so that the fragments can be fused in
    the order they appear in the list by for example Gibson assembly or homologous
    recombination.

    Given that we have two linear :mod:`pydna.amplicon.Amplicon` objects a and b

    we can modify the reverse primer of a and forward primer of b with tails to allow
    fusion by fusion PCR, Gibson assembly or in-vivo homologous recombination.
    The basic requirements for the primers for the three techniques are the same.

    ::



            _________ a _________
           /                     \\
           agcctatcatcttggtctctgca
                             |||||
                            <gacgt
           agcct>
           |||||
           tcggatagtagaaccagagacgt

                                   __________ b ________
                                  /                     \\
                                  TTTATATCGCATGACTCTTCTTT
                                                    |||||
                                                   <AGAAA
                                  TTTAT>
                                  |||||
                                  AAATATAGCGTACTGAGAAGAAA


           agcctatcatcttggtctctgcaTTTATATCGCATGACTCTTCTTT
           ||||||||||||||||||||||||||||||||||||||||||||||
           tcggatagtagaaccagagacgtAAATATAGCGTACTGAGAAGAAA
           \\___________________ c ______________________/


    Design tailed primers incorporating a part of the next or previous fragment to be assembled.

    ::


      agcctatcatcttggtctctgca
      |||||||||||||||||||||||
                      gagacgtAAATATA

      |||||||||||||||||||||||
      tcggatagtagaaccagagacgt


                             TTTATATCGCATGACTCTTCTTT
                             |||||||||||||||||||||||

                      ctctgcaTTTATAT
                             |||||||||||||||||||||||
                             AAATATAGCGTACTGAGAAGAAA

    PCR products with flanking sequences are formed in the PCR process.

    ::

      agcctatcatcttggtctctgcaTTTATAT
      ||||||||||||||||||||||||||||||
      tcggatagtagaaccagagacgtAAATATA
                      \\____________/

                         identical
                         sequences
                       ____________
                      /            \\
                      ctctgcaTTTATATCGCATGACTCTTCTTT
                      ||||||||||||||||||||||||||||||
                      gagacgtAAATATAGCGTACTGAGAAGAAA

    The fragments can be fused by any of the techniques mentioned earlier to form c:

    ::

      agcctatcatcttggtctctgcaTTTATATCGCATGACTCTTCTTT
      ||||||||||||||||||||||||||||||||||||||||||||||
      tcggatagtagaaccagagacgtAAATATAGCGTACTGAGAAGAAA


    The first argument of this function is a list of sequence objects containing
    Amplicons and other similar objects.

    **At least every second sequence object needs to be an Amplicon**

    This rule exists because if a sequence object is that is not a PCR product
    is to be fused with another fragment, that other fragment needs to be an Amplicon
    so that the primer of the other object can be modified to include the whole stretch
    of sequence homology needed for the fusion. See the example below where a is a
    non-amplicon (a linear plasmid  vector for instance)

    ::

       _________ a _________           __________ b ________
      /                     \\         /                     \\
      agcctatcatcttggtctctgca   <-->  TTTATATCGCATGACTCTTCTTT
      |||||||||||||||||||||||         |||||||||||||||||||||||
      tcggatagtagaaccagagacgt                          <AGAAA
                                      TTTAT>
                                      |||||||||||||||||||||||
                                <-->  AAATATAGCGTACTGAGAAGAAA


           agcctatcatcttggtctctgcaTTTATATCGCATGACTCTTCTTT
           ||||||||||||||||||||||||||||||||||||||||||||||
           tcggatagtagaaccagagacgtAAATATAGCGTACTGAGAAGAAA
           \\___________________ c ______________________/


    In this case only the forward primer of b is fitted with a tail with a part a:

    ::


      agcctatcatcttggtctctgca
      |||||||||||||||||||||||
      tcggatagtagaaccagagacgt


                             TTTATATCGCATGACTCTTCTTT
                             |||||||||||||||||||||||
                                              <AGAAA
               tcttggtctctgcaTTTATAT
                             |||||||||||||||||||||||
                             AAATATAGCGTACTGAGAAGAAA

    PCR products with flanking sequences are formed in the PCR process.

    ::

      agcctatcatcttggtctctgcaTTTATAT
      ||||||||||||||||||||||||||||||
      tcggatagtagaaccagagacgtAAATATA
                      \\____________/

                         identical
                         sequences
                       ____________
                      /            \\
                      ctctgcaTTTATATCGCATGACTCTTCTTT
                      ||||||||||||||||||||||||||||||
                      gagacgtAAATATAGCGTACTGAGAAGAAA

    The fragments can be fused by for example Gibson assembly:

    ::

      agcctatcatcttggtctctgcaTTTATAT
      ||||||||||||||||||||||||||||||
      tcggatagtagaacca

                                   TCGCATGACTCTTCTTT
                      ||||||||||||||||||||||||||||||
                      gagacgtAAATATAGCGTACTGAGAAGAAA

    to form c:

    ::

      agcctatcatcttggtctctgcaTTTATATCGCATGACTCTTCTTT
      ||||||||||||||||||||||||||||||||||||||||||||||
      tcggatagtagaaccagagacgtAAATATAGCGTACTGAGAAGAAA


    The first argument of this function is a list of sequence objects containing
    Amplicons and other similar objects.

    The overlap argument controls how many base pairs of overlap required between
    adjacent sequence fragments. In the junction between Amplicons, tails with the
    length of about half of this value is added to the two primers
    closest to the junction.

    ::

            >       <
            Amplicon1
                     Amplicon2
                     >       <

                     ⇣

            >       <-
            Amplicon1
                     Amplicon2
                    ->       <

    In the case of an Amplicon adjacent to a Dseqrecord object, the tail will
    be twice as long (1*overlap) since the
    recombining sequence is present entirely on this primer:

    ::

            Dseqrecd1
                     Amplicon1
                     >       <

                     ⇣

            Dseqrecd1
                     Amplicon1
                   -->       <

    Note that if the sequence of DNA fragments starts or stops with an Amplicon,
    the very first and very last prinmer will not be modified i.e. assembles are
    always assumed to be linear. There are simple tricks around that for circular
    assemblies depicted in the last two examples below.

    The maxlink arguments controls the cut off length for sequences that will be
    synhtesized by adding them to primers for the adjacent fragment(s). The
    argument list may contain short spacers (such as spacers between fusion proteins).


    ::

        Example 1: Linear assembly of PCR products (pydna.amplicon.Amplicon class objects) ------


        >       <         >       <
        Amplicon1         Amplicon3
                 Amplicon2         Amplicon4
                 >       <         >       <

                             ⇣
                             pydna.design.assembly_fragments
                             ⇣

        >       <-       ->       <-                      pydna.assembly.Assembly
        Amplicon1         Amplicon3
                 Amplicon2         Amplicon4     ➤  Amplicon1Amplicon2Amplicon3Amplicon4
                ->       <-       ->       <


        Example 2: Linear assembly of alternating Amplicons and other fragments


        >       <         >       <
        Amplicon1         Amplicon2
                 Dseqrecd1         Dseqrecd2

                             ⇣
                             pydna.design.assembly_fragments
                             ⇣

        >       <--     -->       <--                     pydna.assembly.Assembly
        Amplicon1         Amplicon2
                 Dseqrecd1         Dseqrecd2     ➤  Amplicon1Dseqrecd1Amplicon2Dseqrecd2


        Example 3: Linear assembly of alternating Amplicons and other fragments


        Dseqrecd1         Dseqrecd2
                 Amplicon1         Amplicon2
                 >       <       -->       <

                             ⇣
                     pydna.design.assembly_fragments
                             ⇣
                                                          pydna.assembly.Assembly
        Dseqrecd1         Dseqrecd2
                 Amplicon1         Amplicon2     ➤  Dseqrecd1Amplicon1Dseqrecd2Amplicon2
               -->       <--     -->       <


        Example 4: Circular assembly of alternating Amplicons and other fragments

                         ->       <==
        Dseqrecd1         Amplicon2
                 Amplicon1         Dseqrecd1
               -->       <-
                             ⇣
                             pydna.design.assembly_fragments
                             ⇣
                                                           pydna.assembly.Assembly
                         ->       <==
        Dseqrecd1         Amplicon2                    -Dseqrecd1Amplicon1Amplicon2-
                 Amplicon1                       ➤    |                             |
               -->       <-                            -----------------------------

        ------ Example 5: Circular assembly of Amplicons

        >       <         >       <
        Amplicon1         Amplicon3
                 Amplicon2         Amplicon1
                 >       <         >       <

                             ⇣
                             pydna.design.assembly_fragments
                             ⇣

        >       <=       ->       <-
        Amplicon1         Amplicon3
                 Amplicon2         Amplicon1
                ->       <-       +>       <

                             ⇣
                     make new Amplicon using the Amplicon1.template and
                     the last fwd primer and the first rev primer.
                             ⇣
                                                           pydna.assembly.Assembly
        +>       <=       ->       <-
         Amplicon1         Amplicon3                  -Amplicon1Amplicon2Amplicon3-
                  Amplicon2                      ➤   |                             |
                 ->       <-                          -----------------------------




    Parameters
    ----------

    f : list of :mod:`pydna.amplicon.Amplicon` and other Dseqrecord like objects
        list Amplicon and Dseqrecord object for which fusion primers should be constructed.

    overlap : int, optional
        Length of required overlap between fragments.

    maxlink : int, optional
        Maximum length of spacer sequences that may be present in f. These will be included in tails for designed primers.

    circular : bool, optional
        If True, the assembly is circular. If False, the assembly is linear.

    Returns
    -------
    seqs : list of :mod:`pydna.amplicon.Amplicon` and other Dseqrecord like objects :mod:`pydna.amplicon.Amplicon` objects

        ::

          [Amplicon1,
           Amplicon2, ...]


    Examples
    --------

    >>> from pydna.dseqrecord import Dseqrecord
    >>> from pydna.design import primer_design
    >>> a=primer_design(Dseqrecord("atgactgctaacccttccttggtgttgaacaagatcgacgacatttcgttcgaaacttacgatg"))
    >>> b=primer_design(Dseqrecord("ccaaacccaccaggtaccttatgtaagtacttcaagtcgccagaagacttcttggtcaagttgcc"))
    >>> c=primer_design(Dseqrecord("tgtactggtgctgaaccttgtatcaagttgggtgttgacgccattgccccaggtggtcgtttcgtt"))
    >>> from pydna.design import assembly_fragments
    >>> # We would like a circular recombination, so the first sequence has to be repeated
    >>> fa1,fb,fc,fa2 = assembly_fragments([a,b,c,a])
    >>> # Since all fragments are Amplicons, we need to extract the rp of the 1st and fp of the last fragments.
    >>> from pydna.amplify import pcr
    >>> fa = pcr(fa2.forward_primer, fa1.reverse_primer, a)
    >>> [fa,fb,fc]
    [Amplicon(100), Amplicon(101), Amplicon(102)]
    >>> fa.name, fb.name, fc.name = "fa fb fc".split()
    >>> from pydna.assembly import Assembly
    >>> assemblyobj = Assembly([fa,fb,fc])
    >>> assemblyobj
    Assembly
    fragments....: 100bp 101bp 102bp
    limit(bp)....: 25
    G.nodes......: 6
    algorithm....: common_sub_strings
    >>> assemblyobj.assemble_linear()
    [Contig(-231), Contig(-166), Contig(-36)]
    >>> assemblyobj.assemble_circular()[0].seguid()
    'cdseguid=85t6tfcvWav0wnXEIb-lkUtrl4s'
    >>> (a+b+c).looped().seguid()
    'cdseguid=85t6tfcvWav0wnXEIb-lkUtrl4s'
    >>> print(assemblyobj.assemble_circular()[0].figure())
     -|fa|36
    |     \\/
    |     /\\
    |     36|fb|36
    |           \\/
    |           /\\
    |           36|fc|36
    |                 \\/
    |                 /\\
    |                 36-
    |                    |
     --------------------
    >>>

    """

    # Recursive call for circular assemblies
    if circular:
        fragments = assembly_fragments(
            f + f[0:1], overlap=overlap, maxlink=maxlink, circular=False
        )

        if hasattr(fragments[0], "template"):
            fragments[0] = _pcr(
                (fragments[-1].forward_primer, fragments[0].reverse_primer),
                fragments[0].template,
            )
        return fragments[:-1]

    # sanity check for arguments
    nf = [item for item in f if len(item) > maxlink]
    if not all(
        hasattr(i[0], "template") or hasattr(i[1], "template") for i in zip(nf, nf[1:])
    ):
        raise ValueError(
            "Every second fragment larger than maxlink has to be an Amplicon object"
        )

    # _module_logger.debug("### assembly fragments ###")
    # _module_logger.debug("overlap     = %s", overlap)
    # _module_logger.debug("max_link    = %s", maxlink)

    f = [_copy.copy(f) for f in f]

    first_fragment_length = len(f[0])
    last_fragment_length = len(f[-1])

    if first_fragment_length <= maxlink:
        # first fragment should be removed and added to second fragment (new first fragment) forward primer
        f[1].forward_primer = f[0].seq._data.decode("ASCII") + f[1].forward_primer
        # _module_logger.debug("first fragment removed since len(f[0]) = %s", first_fragment_length)
        f = f[1:]
    # else:
    # _module_logger.debug("first fragment stays since len(f[0]) = %s", first_fragment_length)

    if last_fragment_length <= maxlink:
        f[-2].reverse_primer = (
            f[-1].seq.reverse_complement()._data.decode("ASCII") + f[-2].reverse_primer
        )
        f = f[:-1]
    # _module_logger.debug("last fragment removed since len(f[%s]) = %s", len(f), last_fragment_length)
    # else:
    # _module_logger.debug("last fragment stays since len(f[%s]) = %s", len(f), last_fragment_length)

    empty = _Dseqrecord("")

    # _module_logger.debug(f)
    # _module_logger.debug("loop through fragments in groups of three:")

    tail_length = _math.ceil(overlap / 2)

    for i in range(len(f) - 1):
        first = f[i]
        secnd = f[i + 1]

        secnd_len = len(secnd)

        # _module_logger.debug("first = %s", str(first.seq))
        # _module_logger.debug("secnd = %s", str(secnd.seq))

        if secnd_len <= maxlink:
            # _module_logger.debug("secnd is smaller or equal to maxlink; should be added to primer(s)")
            third = f[i + 2]
            # _module_logger.debug("third = %s", str(third.seq))
            if hasattr(f[i], "template") and hasattr(third, "template"):
                # _module_logger.debug(
                #    "secnd is is flanked by amplicons, so half of secnd should be added each flanking primers"
                # )

                first.reverse_primer = (
                    secnd.seq.reverse_complement()._data.decode("ASCII")[
                        secnd_len // 2 :
                    ]
                    + first.reverse_primer
                )
                third.forward_primer = (
                    secnd.seq._data.decode("ASCII")[secnd_len // 2 :]
                    + third.forward_primer
                )

                lnk = (
                    third.seq.reverse_complement()._data.decode("ASCII")
                    + secnd.reverse_complement().seq._data.decode("ASCII")[
                        : secnd_len // 2
                    ]
                )[-tail_length:]
                # _module_logger.debug("1 %s", lnk)
                first.reverse_primer = lnk + first.reverse_primer

                lnk = (
                    first.seq._data.decode("ASCII")
                    + secnd.seq._data.decode("ASCII")[: secnd_len // 2]
                )[-tail_length:]
                # _module_logger.debug("2 %s", lnk)
                third.forward_primer = lnk + third.forward_primer

            elif hasattr(first, "template"):
                first.reverse_primer = (
                    secnd.seq.reverse_complement()._data.decode("ASCII")
                    + first.reverse_primer
                )
                lnk = str(third.seq[:overlap].reverse_complement())
                first.reverse_primer = lnk + first.reverse_primer
            elif hasattr(third, "template"):
                third.forward_primer = (
                    secnd.seq._data.decode("ASCII") + third.forward_primer
                )
                lnk = str(first.seq[-overlap:])
                third.forward_primer = lnk + third.forward_primer
            secnd = empty
            f[i + 2] = third
        else:  # secnd is larger than maxlink
            if hasattr(first, "template") and hasattr(secnd, "template"):
                lnk = str(first.seq[-tail_length:])
                # #_module_logger.debug("4 %s", lnk)
                secnd.forward_primer = lnk + secnd.forward_primer
                lnk = str(secnd.seq[:tail_length].reverse_complement())
                # #_module_logger.debug("5 %s", lnk)
                first.reverse_primer = lnk + first.reverse_primer
            elif hasattr(first, "template"):
                lnk = str(secnd.seq[:overlap].reverse_complement())
                # #_module_logger.debug("6 %s", lnk)
                first.reverse_primer = lnk + first.reverse_primer
            elif hasattr(secnd, "template"):
                lnk = str(first.seq[-overlap:])
                # #_module_logger.debug("7 %s", lnk)
                secnd.forward_primer = lnk + secnd.forward_primer
        f[i] = first
        f[i + 1] = secnd

    # _module_logger.debug("loop ended")

    f = [item for item in f if len(item)]

    return [
        (
            _pcr(
                p.forward_primer,
                p.reverse_primer,
                p.template,
                limit=min((p.forward_primer._fp, p.reverse_primer._fp)),
            )
            if hasattr(p, "template")
            else p
        )
        for p in f
    ]


def circular_assembly_fragments(f, overlap=35, maxlink=40):
    """
    Equivalent to `assembly_fragments` with `circular=True`.

    Deprecated, kept for backward compatibility. Use `assembly_fragments` with `circular=True` instead.
    """
    import warnings

    warnings.warn(
        "The circular_assembly_fragments function is deprecated. Use assembly_fragments with circular=True instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return assembly_fragments(f, overlap=overlap, maxlink=maxlink, circular=True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2013-2023 by Björn Johansson.  All rights reserved.
# This code is part of the Python-dna distribution and governed by its
# license.  Please see the LICENSE.txt file that should have been included
# as part of this package.
"""Provides a function for downloading online text files."""

import textwrap as _textwrap

# import os as _os
from pydna._pretty import pretty_str as _pretty_str

# from pydna.utils import memorize as _memorize
# import logging as _logging

# _module_logger = _logging.getLogger("pydna." + __name__)


# @_memorize("pydna.download.download_text")
def download_text(url):
    """docstring."""
    import requests

    #    _module_logger.info("#### DOWNLOAD TEXT ####")
    #    _module_logger.info("url = %s", url)
    req = requests.get(url)
    #    _module_logger.info("url = %s", str(req))
    result = _textwrap.dedent(req.text).strip()
    result = result.replace("\r\n", "\n").replace("\r", "\n")
    #    _module_logger.info("result[:160] = %s", result[:160])
    return _pretty_str(result)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# doctest: +NORMALIZE_WHITESPACE
# doctest: +SKIP
# Copyright 2013-2023 by Björn Johansson.  All rights reserved.
# This code is part of the Python-dna distribution and governed by its
# license.  Please see the LICENSE.txt file that should have been included
# as part of this package.

"""docstring."""

import math as _math
from pydna.ladders import GeneRuler_1kb_plus as _mwstd


def interpolator(mwstd):
    """docstring."""
    from scipy.interpolate import CubicSpline

    interpolator = CubicSpline(
        [len(fr) for fr in mwstd[::-1]],
        [fr.rf for fr in mwstd[::-1]],
        bc_type="natural",
        extrapolate=False,
    )
    interpolator.mwstd = mwstd
    return interpolator


def gel(
    samples=None, gel_length=600, margin=50, interpolator=interpolator(mwstd=_mwstd)
):
    import numpy as np
    from PIL import Image as Image
    from PIL import ImageDraw as ImageDraw

    """docstring."""
    max_intensity = 256
    lane_width = 50
    lanesep = 10
    start = 10
    samples = samples or [interpolator.mwstd]
    width = int(60 + (lane_width + lanesep) * len(samples))
    lanes = np.zeros((len(samples), gel_length), dtype=int)
    image = Image.new("RGB", (width, gel_length), "#ddd")
    draw = ImageDraw.Draw(image)
    draw.rectangle((0, 0, (width, gel_length)), fill=(0, 0, 0))
    scale = (gel_length - margin) / interpolator(min(interpolator.x))

    for labelsource in samples[0]:
        peak_centre = (interpolator(len(labelsource))) * scale - 5 + start
        label = f"{len(labelsource):<5} -"
        draw.text((2, peak_centre), label, fill=(255, 255, 255))

    for lane_number, lane in enumerate(samples):
        for band in lane:
            log = _math.log(len(band), 10)
            height = (band.m() / (240 * log)) * 1e10
            peak_centre = interpolator(len(band)) * scale + start
            max_spread = 10
            # if len(band) <= 50:
            #     peak_centre += 50
            #     max_spread *= 4
            #     max_intensity /= 10
            band_spread = max_spread / log
            for i in range(max_spread, 0, -1):
                y1 = peak_centre - i
                y2 = peak_centre + i
                intensity = (
                    height
                    * _math.exp(
                        -float(((y1 - peak_centre) ** 2)) / (2 * (band_spread**2))
                    )
                    * max_intensity
                )
                for y in range(int(y1), int(y2)):
                    try:
                        lanes[lane_number][y] += intensity
                    except IndexError:
                        pass

    for i, lane in enumerate(lanes):
        max_intensity = np.amax(lanes[i])
        if max_intensity > 256:
            lanes[i] = np.multiply(lanes[i], 256)
            lanes[i] = np.divide(lanes[i], max_intensity)

    for i, lane in enumerate(lanes):
        x1 = 50 + i * (lane_width + lanesep)
        x2 = x1 + lane_width
        for y, intensity in enumerate(lane):
            y1 = y
            y2 = y + 1
            draw.rectangle((x1, y1, x2, y2), fill=(intensity, intensity, intensity))

    return image


# Inverting and rotating the gel
# im = gel([ GeneRuler_1kb_plus, [band, ]])
# from PIL import ImageOps
# im_invert = ImageOps.invert(im)
# im.rotate(90, expand=1)



# -*- coding: utf-8 -*-
"""conftest.py"""

# import pathlib
# import os

# cwd = pathlib.Path(__file__).parent

# def pytest_runtest_setup(item):
#     # called for running each test in 'a' directory
#     #os.chdir(cwd)
#     pass


# def pytest_configure(config):
#    """
#    Allows plugins and conftest files to perform initial configuration.
#    This hook is called for every plugin and initial conftest
#    file after command line options have been parsed.
#    """
#    os.chdir(cwd)
#    print(f"cwd set to {cwd} in {__file__}")


# def pytest_sessionstart(session):
#     """
#     Called after the Session object has been created and
#     before performing collection and entering the run test loop.
#     """


# def pytest_sessionfinish(session, exitstatus):
#     """
#     Called after whole test run finished, right before
#     returning the exit status to the system.
#     """


# def pytest_unconfigure(config):
#     """
#     called before test process is exited.
#     """



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Classes for nicer Jupyter output.

The pretty_str class is similar to str but has a _repr_pretty_ method
for for nicer string output in the IPython shell and Jupyter notebook.
"""

from prettytable import PrettyTable as _Pt
from prettytable import TableStyle as _TableStyle
from copy import copy as _copy
from typing import List as _List


class pretty_str(str):
    """Thanks to Min RK, UC Berkeley for this."""

    def _repr_pretty_(self, p, cycle):
        p.text(self)


class PrettyTable(_Pt):
    """docstring."""

    def lol(self) -> _List[list]:
        """docstring."""
        return [self._field_names] + self._rows

    def __repr__(self) -> str:
        """docstring."""
        return self.get_string()

    def _repr_markdown_(self) -> pretty_str:
        c = _copy(self)
        c.set_style(_TableStyle.MARKDOWN)
        return pretty_str(c.get_string())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2013-2023 by Björn Johansson.  All rights reserved.
# This code is part of the Python-dna distribution and governed by its
# license.  Please see the LICENSE.txt file that should have been included
# as part of this package.

from pydna.dseqrecord import Dseqrecord
import os as _os

# import logging as _logging
from Bio.Blast import NCBIWWW
from Bio.Blast import NCBIXML

# _module_logger = _logging.getLogger("pydna." + __name__)


email = _os.getenv("pydna_email")
tool = "pydna"


def genbank_accession(s: str) -> Dseqrecord:
    """docstring."""
    s = Dseqrecord(s)

    NCBIWWW.email = email
    NCBIWWW.tool = tool

    result_handle = NCBIWWW.qblast(
        "blastn",
        "nt",
        str(s.seq),
        hitlist_size=1,
        alignments=1,
        descriptions=1,
        expect=1e-8,
        megablast=True,
        service="megablast",
        ungapped_alignment=True,
    )

    blast_records = NCBIXML.read(result_handle)
    best_alignment, *rest = blast_records.alignments
    best_hsp, *rest = best_alignment.hsps
    dbs = best_hsp.sbjct
    start, stop = sorted((best_hsp.sbjct_start, best_hsp.sbjct_end))
    result = Dseqrecord(
        dbs,
        circular=False,
        id=s.name,
        name=s.name,
        description=(f"{best_alignment.accession} " f"REGION: {start}..{stop}"),
    )
    return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2013-2023 by Björn Johansson.  All rights reserved.
# This code is part of the Python-dna distribution and governed by its
# license.  Please see the LICENSE.txt file that should have been included
# as part of this package.
from pydna.dseqrecord import Dseqrecord as _Dseqrecord


class GenbankFile(_Dseqrecord):
    def __init__(self, record, *args, path=None, **kwargs):
        super().__init__(record, *args, **kwargs)
        self.path = path

    @classmethod
    def from_SeqRecord(cls, record, *args, path=None, **kwargs):
        obj = super().from_SeqRecord(record, *args, **kwargs)
        obj.path = path
        return obj

    def __repr__(self):
        """returns a short string representation of the object"""
        return "File({})({}{})".format(
            self.id, {True: "-", False: "o"}[not self.circular], len(self)
        )

    def _repr_pretty_(self, p, cycle):
        """returns a short string representation of the object"""
        p.text(
            "File({})({}{})".format(
                self.id, {True: "-", False: "o"}[not self.circular], len(self)
            )
        )

    def _repr_html_(self):
        return "<a href='{path}' target='_blank'>{path}</a><br>".format(path=self.path)

    def reverse_complement(self):
        answer = type(self)(super().reverse_complement(), path=self.path)
        return answer

    rc = reverse_complement



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2013-2023 by Björn Johansson.  All rights reserved.
# This code is part of the Python-dna distribution and governed by its
# license.  Please see the LICENSE.txt file that should have been included
# as part of this package.
"""Miscellaneous functions."""

from Bio.Data.IUPACData import ambiguous_dna_complement as _ambiguous_dna_complement
from Bio.Seq import _maketrans

# import shelve as _shelve
# import os as _os
import re as _re

# import logging as _logging
# import base64 as _base64
# import pickle as _pickle
# import hashlib as _hashlib
import keyword as _keyword
import collections as _collections
import itertools as _itertools
from copy import deepcopy as _deepcopy

import sys as _sys
import random
import subprocess as _subprocess
from bisect import bisect as _bisect
from math import ceil as _ceil

from pydna.codon import weights as _weights
from pydna.codon import rare_codons as _rare_codons

from Bio.SeqFeature import SimpleLocation as _sl
from Bio.SeqFeature import CompoundLocation as _cl
from Bio.SeqFeature import Location as _Location

from typing import Union as _Union, TypeVar as _TypeVar, List as _List

# For functions that take str or bytes as input and return str or bytes as output, matching the input type
StrOrBytes = _TypeVar("StrOrBytes", str, bytes)

# _module_logger = _logging.getLogger("pydna." + __name__)
_ambiguous_dna_complement.update({"U": "A"})
_complement_table = _maketrans(_ambiguous_dna_complement)


def three_frame_orfs(
    dna: str,
    limit: int = 100,
    startcodons: tuple = ("ATG",),
    stopcodons: tuple = ("TAG", "TAA", "TGA"),
    # startcodons: tuple[str, ...] = ("ATG",),
    # stopcodons: tuple[str, ...] = ("TAG", "TAA", "TGA"),
):
    """Overlapping orfs in three frames."""
    # breakpoint()
    limit = _ceil(limit / 3) - 1
    dna = dna.upper()

    orfs = []

    for frame in (0, 1, 2):

        codons = [dna[i : i + 3] for i in range(frame, len(dna), 3)]

        startdindices = [i for i, cd in enumerate(codons) if cd in startcodons]
        stopdindices = [i for i, cd in enumerate(codons) if cd in stopcodons]

        for startindex in startdindices:
            try:
                stopindex = stopdindices[_bisect(stopdindices, startindex)]
            except IndexError:
                pass
            else:
                if stopindex - startindex >= limit:
                    orfs.append(
                        (frame, startindex * 3 + frame, (stopindex + 1) * 3 + frame)
                    )
                # print(stopindex, startindex, limit)
    return orfs


def shift_location(original_location, shift, lim):
    """docstring."""
    newparts = []
    strand = original_location.strand
    if lim is None:
        if min(original_location) + shift < 0:
            raise ValueError(
                "Shift moves location below zero, use a `lim` to loop around if sequence is circular."
            )
        lim = _sys.maxsize

    for part in original_location.parts:
        new_start = (part.start + shift) % lim
        new_end = (part.end + shift) % lim or lim
        old_start, old_end = (
            (newparts[-1].start, newparts[-1].end) if len(newparts) else (None, None)
        )

        # The "join with old" cases are for features with multiple parts
        # in which consecutive parts do not have any bases between them.
        # This type of feature is generated to represent a feature that
        # spans the origin of a circular sequence. See more details in
        # https://github.com/pydna-group/pydna/issues/195

        if len(part) == 0:
            newparts.append(_sl(new_start, new_start, strand))
            continue
        # Join with old, case 1
        elif strand != -1 and old_end == new_start:
            part = newparts.pop()
            part._end = new_end
            new_start = part.start
        # Join with old, case 2
        elif strand == -1 and old_start == new_end:
            part = newparts.pop()
            part._start = new_start
            new_end = part.end
        if new_start < new_end:
            newparts.append(_sl(new_start, new_end, strand))
        else:
            parttuple = (_sl(new_start, lim, strand), _sl(0, new_end, strand))
            newparts.extend(parttuple if strand != -1 else parttuple[::-1])
    try:
        newloc = _cl(newparts)
    except ValueError:
        newloc, *n = newparts
    assert len(newloc) == len(original_location)
    return newloc


# def shift_feature(feature, shift, lim):
#     """Return a new feature with shifted location."""
#     # TODO: Missing tests
#     new_location = shift_location(feature.location, shift, lim)
#     new_feature = _deepcopy(feature)
#     new_feature.location = new_location
#     return new_feature


def shift_feature(feature, shift, lim):
    """Return a new feature with shifted location."""
    # TODO: Missing tests
    new_location = shift_location(feature.location, shift, lim)
    new_feature = _deepcopy(feature)
    new_feature.location = new_location
    return new_feature


# def smallest_rotation(s):
#     """Smallest rotation of a string.

#     Algorithm described in Pierre Duval, Jean. 1983. Factorizing Words
#     over an Ordered Alphabet. Journal of Algorithms & Computational Technology
#     4 (4) (December 1): 363–381. and Algorithms on strings and sequences based
#     on Lyndon words, David Eppstein 2011.
#     https://gist.github.com/dvberkel/1950267

#     Examples
#     --------
#     >>> from pydna.utils import smallest_rotation
#     >>> smallest_rotation("taaa")
#     'aaat'

#     """
#     prev, rep = None, 0
#     ds = _array("u", 2 * s)
#     lens, lends = len(s), len(ds)
#     old = 0
#     k = 0
#     w = ""
#     while k < lends:
#         i, j = k, k + 1
#         while j < lends and ds[i] <= ds[j]:
#             i = (ds[i] == ds[j]) and i + 1 or k
#             j += 1
#         while k < i + 1:
#             k += j - i
#             prev = w
#             w = ds[old:k]
#             old = k
#             if w == prev:
#                 rep += 1
#             else:
#                 prev, rep = w, 1
#             if len(w) * rep == lens:
#                 return "".join(w * rep)


def smallest_rotation(s):
    """Smallest rotation of a string.

    Algorithm described in Pierre Duval, Jean. 1983. Factorizing Words
    over an Ordered Alphabet. Journal of Algorithms & Computational Technology
    4 (4) (December 1): 363–381. and Algorithms on strings and sequences based
    on Lyndon words, David Eppstein 2011.
    https://gist.github.com/dvberkel/1950267

    Examples
    --------
    >>> from pydna.utils import smallest_rotation
    >>> smallest_rotation("taaa")
    'aaat'
    """
    from pydivsufsort import min_rotation

    k = min_rotation(bytes(s, "ascii"))
    return s[k:] + s[:k]


def cai(seq: str, organism: str = "sce", weights: dict = _weights):
    """docstring."""
    from cai2 import CAI as _CAI

    return round(_CAI(seq.upper(), weights=weights[organism]), 3)


def rarecodons(seq: str, organism="sce"):
    """docstring."""
    rare = _rare_codons[organism]
    s = seq.upper()
    slices = []
    for i in range(0, len(seq) // 3):
        x, y = i * 3, i * 3 + 3
        trip = s[x:y]
        if trip in rare:
            slices.append(slice(x, y, 1))
    return slices


def express(seq: str, organism="sce"):
    """docstring.

    **NOT IMPLEMENTED YET**
    """
    # x = _PrettyTable(["cds", "len", "cai", "gc", "sta", "stp", "n-end"] + _rare_codons[organism] + ["rare"])
    # val = []

    # val.append(f"{self._data.upper().decode('ASCII')[:3]}..." f"{self._data.upper().decode('ASCII')[-3:]}")
    # val.append(len(self) / 3)
    # val.append(cai(organism))
    # val.append(gc())
    # val.append(startcodon())
    # val.append(stopcodon())
    # val.append(_n_end[organism].get(_seq3(self[3:6].translate())))
    # s = self._data.upper().decode("ASCII")
    # trps = [s[i * 3 : i * 3 + 3] for i in range(0, len(s) // 3)]
    # tot = 0
    # for cdn in _rare_codons[organism]:
    #     cnt = trps.count(cdn)
    #     tot += cnt
    #     val.append(cnt)
    # val.append(round(tot / len(trps), 3))
    # x.add_row(val)
    # return x
    raise NotImplementedError


def open_folder(pth):
    """docstring."""
    if _sys.platform == "win32":
        _subprocess.run(["start", pth], shell=True)
    elif _sys.platform == "darwin":
        _subprocess.run(["open", pth])
    else:
        try:
            _subprocess.run(["xdg-open", pth])
        except OSError:
            return "no cache to open."


def rc(sequence: StrOrBytes) -> StrOrBytes:
    """Reverse complement.

    accepts mixed DNA/RNA
    """
    return sequence.translate(_complement_table)[::-1]


def complement(sequence: str):
    """Complement.

    accepts mixed DNA/RNA
    """
    return sequence.translate(_complement_table)


# def memorize(filename):
#     """Cache functions and classes.

#     see pydna.download
#     """

#     def decorator(f):
#         def wrappee(*args, **kwargs):
#             _module_logger.info("#### memorizer ####")
#             _module_logger.info("cache filename                   = %s", filename)
#             _module_logger.info(
#                 "os.environ['pydna_cached_funcs'] = %s",
#                 _os.getenv("pydna_cached_funcs", ""),
#             )
#             if filename not in _os.getenv("pydna_cached_funcs", ""):
#                 _module_logger.info("cache filename not among cached functions, made it new!")
#                 return f(*args, **kwargs)
#             key = _base64.urlsafe_b64encode(_hashlib.sha1(_pickle.dumps((args, kwargs))).digest()).decode("ascii")
#             _module_logger.info("key = %s", key)
#             cache = _shelve.open(
#                 _os.path.join(_os.environ["pydna_data_dir"], identifier_from_string(filename)),
#                 writeback=False,
#             )
#             try:
#                 result = cache[key]
#             except KeyError:
#                 _module_logger.info(
#                     "no result for key %s in shelve %s",
#                     key,
#                     identifier_from_string(filename),
#                 )
#                 result = f(*args, **kwargs)
#                 _module_logger.info("made it new!")
#                 cache[key] = result
#                 _module_logger.info("saved result under key %s", key)
#             else:
#                 _module_logger.info("found %s in cache", key)
#             cache.close()
#             return result

#         return wrappee

#     return decorator


def identifier_from_string(s: str) -> str:
    """Return a valid python identifier.

    based on the argument s or an empty string
    """
    s = s.strip()
    s = _re.sub(r"\s+", r"_", s)
    s.replace("-", "_")
    s = _re.sub("[^0-9a-zA-Z_]", "", s)
    if s and not s[0].isidentifier() or _keyword.iskeyword(s):
        s = "_{s}".format(s=s)
    assert s == "" or s.isidentifier()
    return s


def flatten(*args) -> _List:
    """Flattens an iterable of iterables.

    Down to str, bytes, bytearray or any of the pydna or Biopython seq objects
    """
    output = []
    args = list(args)
    while args:
        top = args.pop()
        if (
            isinstance(top, _collections.abc.Iterable)
            and not isinstance(top, (str, bytes, bytearray))
            and not hasattr(top, "reverse_complement")
        ):
            args.extend(top)
        else:
            output.append(top)
    return output[::-1]


def seq31(seq):
    """Turn a three letter code protein sequence into one with one letter code.

    The single input argument 'seq' should be a protein sequence using single
    letter codes, as a python string.

    This function returns the amino acid sequence as a string using the one
    letter amino acid codes. Output follows the IUPAC standard (including
    ambiguous characters B for "Asx", J for "Xle" and X for "Xaa", and also U
    for "Sel" and O for "Pyl") plus "Ter" for a terminator given as an
    asterisk.

    Any unknown
    character (including possible gap characters), is changed into 'Xaa'.

    Examples
    --------
    >>> from Bio.SeqUtils import seq3
    >>> seq3("MAIVMGRWKGAR*")
    'MetAlaIleValMetGlyArgTrpLysGlyAlaArgTer'
    >>> from pydna.utils import seq31
    >>> seq31('MetAlaIleValMetGlyArgTrpLysGlyAlaArgTer')
    'M  A  I  V  M  G  R  W  K  G  A  R  *'
    """
    threecode = {
        "Ala": "A",
        "Asx": "B",
        "Cys": "C",
        "Asp": "D",
        "Glu": "E",
        "Phe": "F",
        "Gly": "G",
        "His": "H",
        "Ile": "I",
        "Lys": "K",
        "Leu": "L",
        "Met": "M",
        "Asn": "N",
        "Pro": "P",
        "Gln": "Q",
        "Arg": "R",
        "Ser": "S",
        "Thr": "T",
        "Val": "V",
        "Trp": "W",
        "Tyr": "Y",
        "Glx": "Z",
        "Xaa": "X",
        "Ter": "*",
        "Sel": "U",
        "Pyl": "O",
        "Xle": "J",
    }

    nr_of_codons = int(len(seq) / 3)
    sequence = [seq[i * 3 : i * 3 + 3].title() for i in range(nr_of_codons)]
    padding = " " * 2
    return padding.join([threecode.get(aa, "X") for aa in sequence])


def randomRNA(length, maxlength=None):
    """docstring."""
    if maxlength and maxlength > length:
        length = int(round(random.triangular(length, maxlength)))
    return "".join([random.choice("GAUC") for x in range(length)])


def randomDNA(length, maxlength=None):
    """docstring."""
    if maxlength and maxlength > length:
        length = int(round(random.triangular(length, maxlength)))
    return "".join([random.choice("GATC") for x in range(length)])


def randomORF(length, maxlength=None):
    """docstring."""
    length -= 2
    if maxlength and maxlength > length:
        length = int(round(random.triangular(length, maxlength - 2)))

    cdns = (
        "TTT",
        "TTC",
        "TTA",
        "TTG",
        "TCT",
        "TCC",
        "TCA",
        "TCG",
        "TAT",
        "TAC",
        "TGT",
        "TGC",
        "TGG",
        "CTT",
        "CTC",
        "CTA",
        "CTG",
        "CCT",
        "CCC",
        "CCA",
        "CCG",
        "CAT",
        "CAC",
        "CAA",
        "CAG",
        "CGT",
        "CGC",
        "CGA",
        "CGG",
        "ATT",
        "ATC",
        "ATA",
        "ATG",
        "ACT",
        "ACC",
        "ACA",
        "ACG",
        "AAT",
        "AAC",
        "AAA",
        "AAG",
        "AGT",
        "AGC",
        "AGA",
        "AGG",
        "GTT",
        "GTC",
        "GTA",
        "GTG",
        "GCT",
        "GCC",
        "GCA",
        "GCG",
        "GAT",
        "GAC",
        "GAA",
        "GAG",
        "GGT",
        "GGC",
        "GGA",
        "GGG",
    )

    starts = ("ATG",)
    stops = ("TAA", "TAG", "TGA")

    return (
        random.choice(starts)
        + "".join([random.choice(cdns) for x in range(length)])
        + random.choice(stops)
    )


def randomprot(length, maxlength=None):
    """docstring."""
    if maxlength and maxlength > length:
        length = int(round(random.triangular(length, maxlength)))
    return "".join([random.choice("ACDEFGHIKLMNPQRSTVWY") for x in range(length)])


def eq(*args, **kwargs):
    """Compare two or more DNA sequences for equality.

    Compares two or more DNA sequences for equality i.e. if they
    represent the same double stranded DNA molecule.

    Parameters
    ----------
    args : iterable
        iterable containing sequences
        args can be strings, Biopython Seq or SeqRecord, Dseqrecord
        or dsDNA objects.
    circular : bool, optional
        Consider all molecules circular or linear
    linear : bool, optional
        Consider all molecules circular or linear

    Returns
    -------
    eq : bool
        Returns True or False

    Notes
    -----
    Compares two or more DNA sequences for equality i.e. if they
    represent the same DNA molecule.

    Two linear sequences are considiered equal if either:

    1. They have the same sequence (case insensitive)
    2. One sequence is the reverse complement of the other

    Two circular sequences are considered equal if they are circular
    permutations meaning that they have the same length and:

    1. One sequence can be found in the concatenation of the other sequence with itself.
    2. The reverse complement of one sequence can be found in the concatenation of the other sequence with itself.

    The topology for the comparison can be set using one of the keywords
    linear or circular to True or False.

    If circular or linear is not set, it will be deduced from the topology of
    each sequence for sequences that have a linear or circular attribute
    (like Dseq and Dseqrecord).

    Examples
    --------
    >>> from pydna.dseqrecord import Dseqrecord
    >>> from pydna.utils import eq
    >>> eq("aaa","AAA")
    True
    >>> eq("aaa","AAA","TTT")
    True
    >>> eq("aaa","AAA","TTT","tTt")
    True
    >>> eq("aaa","AAA","TTT","tTt", linear=True)
    True
    >>> eq("Taaa","aTaa", linear = True)
    False
    >>> eq("Taaa","aTaa", circular = True)
    True
    >>> a=Dseqrecord("Taaa")
    >>> b=Dseqrecord("aTaa")
    >>> eq(a,b)
    False
    >>> eq(a,b,circular=True)
    True
    >>> a=a.looped()
    >>> b=b.looped()
    >>> eq(a,b)
    True
    >>> eq(a,b,circular=False)
    False
    >>> eq(a,b,linear=True)
    False
    >>> eq(a,b,linear=False)
    True
    >>> eq("ggatcc","GGATCC")
    True
    >>> eq("ggatcca","GGATCCa")
    True
    >>> eq("ggatcca","tGGATCC")
    True
    """
    args = flatten(args)  # flatten

    topology = None

    if "linear" in kwargs:
        if kwargs["linear"] is True:
            topology = "linear"
        if kwargs["linear"] is False:
            topology = "circular"
    elif "circular" in kwargs:
        if kwargs["circular"] is True:
            topology = "circular"
        if kwargs["circular"] is False:
            topology = "linear"
    else:
        topology = set(
            [arg.circular if hasattr(arg, "circular") else None for arg in args]
        )

        if len(topology) != 1:
            raise ValueError("sequences have different topologies")
        topology = topology.pop()
        if topology in (False, None):
            topology = "linear"
        elif topology is True:
            topology = "circular"

    args = [arg.seq if hasattr(arg, "seq") else arg for arg in args]
    args_string_list = [
        arg.watson.lower() if hasattr(arg, "watson") else str(arg).lower()
        for arg in args
    ]

    length = set((len(s) for s in args_string_list))

    if len(length) != 1:
        return False
    same = True

    if topology == "circular":
        # force circular comparison of all given sequences
        for s1, s2 in _itertools.combinations(args_string_list, 2):
            if not (s1 in s2 + s2 or rc(s1) in s2 + s2):
                same = False
    elif topology == "linear":
        # force linear comparison of all given sequences
        for s1, s2 in _itertools.combinations(args_string_list, 2):
            if not (s1 == s2 or s1 == rc(s2)):
                same = False
    return same


# def cuts_overlap(left_cut, right_cut, seq_len):
#     # Special cases:
#     if left_cut is None or right_cut is None or left_cut == right_cut:
#         return False

#     # This block of code would not be necessary if the cuts were
#     # initially represented like this
#     (left_watson, left_ovhg), _ = left_cut
#     (right_watson, right_ovhg), _ = right_cut
#     # Position of the cut on the crick strands on the left and right
#     left_crick = left_watson - left_ovhg
#     right_crick = right_watson - right_ovhg
#     if left_crick >= seq_len:
#         left_crick -= seq_len
#         left_watson -= seq_len
#     if right_crick >= seq_len:
#         right_crick -= seq_len
#         right_watson -= seq_len

#     # Convert into ranges x and y and see if ranges overlap
#     x = sorted([left_watson, left_crick])
#     y = sorted([right_watson, right_crick])
#     return (x[1] > y[0]) != (y[1] < x[0])


# def location_boundaries(loc: _Union[_sl, _cl]):
#     if loc.strand == -1:
#         return loc.parts[-1].start, loc.parts[0].end
#     else:
#         return loc.parts[0].start, loc.parts[-1].end


def cuts_overlap(left_cut, right_cut, seq_len):
    # Special cases:
    if left_cut is None or right_cut is None or left_cut == right_cut:
        return False

    # This block of code would not be necessary if the cuts were
    # initially represented like this
    (left_watson, left_ovhg), _ = left_cut
    (right_watson, right_ovhg), _ = right_cut
    # Position of the cut on the crick strands on the left and right
    left_crick = left_watson - left_ovhg
    right_crick = right_watson - right_ovhg
    if left_crick >= seq_len:
        left_crick -= seq_len
        left_watson -= seq_len
    if right_crick >= seq_len:
        right_crick -= seq_len
        right_watson -= seq_len

    # Convert into ranges x and y and see if ranges overlap
    x = sorted([left_watson, left_crick])
    y = sorted([right_watson, right_crick])
    return (x[1] > y[0]) != (y[1] < x[0])


def location_boundaries(loc: _Union[_sl, _cl]):
    if loc.strand == -1:
        return loc.parts[-1].start, loc.parts[0].end
    else:
        return loc.parts[0].start, loc.parts[-1].end


def locations_overlap(loc1: _Union[_sl, _cl], loc2: _Union[_sl, _cl], seq_len):
    start1, end1 = location_boundaries(loc1)
    start2, end2 = location_boundaries(loc2)

    boundaries1 = [(start1, end1)]
    boundaries2 = [(start2, end2)]

    if start1 > end1:
        boundaries1 = [
            [start1, end1 + seq_len],
            [start1 - seq_len, end1],
        ]
    if start2 > end2:
        boundaries2 = [
            [start2, end2 + seq_len],
            [start2 - seq_len, end2],
        ]

    for b1, b2 in _itertools.product(boundaries1, boundaries2):
        if b1[0] < b2[1] and b1[1] > b2[0]:
            return True

    return False


def sum_is_sticky(
    three_prime_end: tuple[str, str],
    five_prime_end: tuple[str, str],
    partial: bool = False,
) -> int:
    """Return the overlap length if the 3' end of seq1 and 5' end of seq2 ends are sticky and compatible for ligation.
    Return 0 if they are not compatible."""
    type_seq1, sticky_seq1 = three_prime_end
    type_seq2, sticky_seq2 = five_prime_end

    if (
        "blunt" != type_seq2
        and type_seq2 == type_seq1
        and str(sticky_seq2) == str(rc(sticky_seq1))
    ):
        return len(sticky_seq1)

    if not partial:
        return 0

    if type_seq1 != type_seq2 or type_seq2 == "blunt":
        return 0
    elif type_seq2 == "5'":
        sticky_seq1 = str(rc(sticky_seq1))
    elif type_seq2 == "3'":
        sticky_seq2 = str(rc(sticky_seq2))

    ovhg_len = min(len(sticky_seq1), len(sticky_seq2))
    # [::-1] to try the longest overhangs first
    for i in range(1, ovhg_len + 1)[::-1]:
        if sticky_seq1[-i:] == sticky_seq2[:i]:
            return i
    else:
        return 0


def limit_iterator(iterator, limit):
    """
    Call the function with an iterator to raise an error if the number of items is greater than the limit.
    """
    for i, x in enumerate(iterator):
        if i >= limit:
            raise ValueError(f"Too many possible paths (more than {limit})")
        yield x


def create_location(
    start: int, end: int, lim: int, strand: int | None = None
) -> _Location:
    """
    Create a location object from a start and end position.
    If the end position is less than the start position, the location is circular. It handles negative positions.

    Parameters
    ----------
    start : int
        The start position of the location.
    end : int
        The end position of the location.
    lim : int
        The length of the sequence.
    strand : int, optional
        The strand of the location. None, 1 or -1.

    Returns
    -------
    location : Location
        The location object. Can be a SimpleLocation or a CompoundLocation if the feature spans the origin of
        a circular sequence.

    Examples
    --------
    >>> from pydna.utils import create_location
    >>> str(create_location(0, 5, 10,-1))
    '[0:5](-)'
    >>> str(create_location(0, 5, 10,+1))
    '[0:5](+)'
    >>> str(create_location(0, 5, 10))
    '[0:5]'
    >>> str(create_location(8, 2, 10))
    'join{[8:10], [0:2]}'
    >>> str(create_location(8, 2, 10,-1))
    'join{[0:2](-), [8:10](-)}'
    >>> str(create_location(-2, 2, 10))
    'join{[8:10], [0:2]}'

    Note this special case, 0 is the same as len(seq)
    >>> str(create_location(5, 0, 10))
    '[5:10]'

    Note the special case where if start and end are the same,
    the location spans the entire sequence (it's not empty).
    >>> str(create_location(5, 5, 10))
    'join{[5:10], [0:5]}'

    """
    while start < 0:
        start += lim
    while end < 0:
        end += lim
    if end > start:
        return _sl(start, end, strand)
    else:
        return shift_location(_sl(start, end + lim, strand), 0, lim)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2013-2023 by Björn Johansson.  All rights reserved.
# This code is part of the Python-dna distribution and governed by its
# license.  Please see the LICENSE.txt file that should have been included
# as part of this package.
"""docstring."""


class FakeSeq:
    """docstring."""

    def __init__(
        self,
        length: int,
        n: float = 50e-15,  # 50 fmol = 0.05 pmol
        rf: float = 0.0,
    ) -> None:
        self._length = length
        self.n = n
        self.rf = rf

    def m(self) -> float:
        """Mass of the DNA molecule in grams."""
        # M(Da) * n (mol) = g
        return self.M() * self.n

    def M(self) -> float:
        """M grams/mol."""
        return (308.9 * self._length + 79.0) * 2

    def __len__(self) -> int:
        """docstring."""
        return self._length

    def __lt__(self, other) -> bool:
        """docstring."""
        return self._length < len(other)

    def __repr__(self) -> str:
        """docstring."""
        return f"FakeSeq({self._length:.1e})"

    def __str__(self) -> str:
        """docstring."""
        return self.__repr__()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2013-2023 by Björn Johansson.  All rights reserved.
# This code is part of the Python-dna distribution and governed by its
# license.  Please see the LICENSE.txt file that should have been included
# as part of this package.


# from abc import ABC, abstractmethod
# import re
# from pydna.utils import rc


"""
Nicking & USER cloning


CGAuGTCGACTTAGATCTCACAGGCTTTTTTCAAGaCGGCCTTGAATTCAGTCATTTGGATCCGGCCGATC
GCTACAGCTGAATCTAGAGTGTCCGAAAAAAGTTCTGCCGGAACTTAAGTCAGTAAACCTAGGCCGGCuAG

1. Digest both strands
2. Collect all linear ssDNA
3. Anneal all combinations
4. Keep ones present in original molecule
5. Rank by stability
6.

"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2013-2023 by Björn Johansson.  All rights reserved.
# This code is part of the Python-dna distribution and governed by its
# license.  Please see the LICENSE.txt file that should have been included
# as part of this package.

"""This module is based on the Py-rstr-max package that
was written by Romain Brixtel (rbrixtel_at_gmail_dot_com)
(https://brixtel.users.greyc.fr) and is available from
https://code.google.com/p/py-rstr-max
https://github.com/gip0/py-rstr-max
the original code was covered by an MIT licence."""


# from array import array as _array
# import itertools as _itertools
from operator import itemgetter as _itemgetter
from typing import List as _List, Tuple as _Tuple

Match = _Tuple[int, int, int]  # (x_start, y_start, length)

# def _kark_sort(s, SA, n, K):
#     def radixpass(a, b, r, s, n, k):
#         c = _array("i", [0] * (k + 1))
#         for i in range(n):
#             c[r[a[i] + s]] += 1

#         somme = 0
#         for i in range(k + 1):
#             freq, c[i] = c[i], somme
#             somme += freq

#         for i in range(n):
#             b[c[r[a[i] + s]]] = a[i]
#             c[r[a[i] + s]] += 1

#     n0 = (n + 2) // 3
#     n1 = (n + 1) // 3
#     n2 = n // 3
#     n02 = n0 + n2

#     SA12 = _array("i", [0] * (n02 + 3))
#     SA0 = _array("i", [0] * n0)

#     s12 = [i for i in range(n + (n0 - n1)) if i % 3]
#     s12.extend([0] * 3)
#     s12 = _array("i", s12)

#     radixpass(s12, SA12, s, 2, n02, K)
#     radixpass(SA12, s12, s, 1, n02, K)
#     radixpass(s12, SA12, s, 0, n02, K)

#     name = 0
#     c0, c1, c2 = -1, -1, -1
#     for i in range(n02):
#         if s[SA12[i]] != c0 or s[SA12[i] + 1] != c1 or s[SA12[i] + 2] != c2:
#             name += 1
#             c0 = s[SA12[i]]
#             c1 = s[SA12[i] + 1]
#             c2 = s[SA12[i] + 2]
#         if SA12[i] % 3 == 1:
#             s12[SA12[i] // 3] = name
#         else:
#             s12[SA12[i] // 3 + n0] = name

#     if name < n02:
#         _kark_sort(s12, SA12, n02, name + 1)
#         for i in range(n02):
#             s12[SA12[i]] = i + 1
#     else:
#         for i in range(n02):
#             SA12[s12[i] - 1] = i

#     s0 = _array("i", [SA12[i] * 3 for i in range(n02) if SA12[i] < n0])
#     radixpass(s0, SA0, s, 0, n0, K)

#     p = j = k = 0
#     t = n0 - n1
#     while k < n:
#         i = SA12[t] * 3 + 1 if SA12[t] < n0 else (SA12[t] - n0) * 3 + 2
#         j = SA0[p] if p < n0 else 0

#         if SA12[t] < n0:
#             test = (s12[SA12[t] + n0] <= s12[j // 3]) if (s[i] == s[j]) else (s[i] < s[j])
#         elif s[i] == s[j]:
#             test = s12[SA12[t] - n0 + 1] <= s12[j // 3 + n0] if (s[i + 1] == s[j + 1]) else s[i + 1] < s[j + 1]
#         else:
#             test = s[i] < s[j]
#         if test:
#             SA[k] = i
#             t += 1
#             if t == n02:
#                 k += 1
#                 while p < n0:
#                     SA[k] = SA0[p]
#                     p += 1
#                     k += 1
#         else:
#             SA[k] = j
#             p += 1
#             if p == n0:
#                 k += 1
#                 while t < n02:
#                     SA[k] = (SA12[t] * 3) + 1 if SA12[t] < n0 else ((SA12[t] - n0) * 3) + 2
#                     t += 1
#                     k += 1
#         k += 1


# class _Rstr_max:
#     def __init__(self):
#         self.array_str = []

#     def add_str(self, str_unicode):
#         self.array_str.append(str_unicode)

#     def _step1_sort_suffix(self):
#         char_frontier = chr(2)

#         self.global_suffix = char_frontier.join(self.array_str)

#         nbChars = len(self.global_suffix)
#         init = [-1] * nbChars
#         self.idxString = _array("i", init)
#         self.idxPos = _array("i", init)
#         self.endAt = _array("i", init)

#         k = idx = 0
#         for mot in self.array_str:
#             last = k + len(mot)
#             for p in range(len(mot)):
#                 self.idxString[k] = idx
#                 self.idxPos[k] = p
#                 self.endAt[k] = last
#                 k += 1
#             idx += 1
#             k += 1

#         s = self.global_suffix
#         alphabet = [None] + sorted(set(s))
#         k = len(alphabet)
#         n = len(s)
#         t = dict((c, i) for i, c in enumerate(alphabet))
#         SA = _array("i", [0] * (n + 3))
#         _kark_sort(_array("i", [t[c] for c in s] + [0] * 3), SA, n, k)
#         self.res = SA[:n]

#     def _step2_lcp(self):
#         n = len(self.res)
#         init = [0] * n
#         rank = _array("i", init)
#         LCP = _array("i", init)

#         s = self.global_suffix
#         suffix__array = self.res
#         endAt = self.endAt

#         for i in range(len(self.array_str), n):
#             v = self.res[i]
#             rank[v] = i

#         l = 0
#         for j in range(n):
#             if l > 0:
#                 l -= 1
#             i = rank[j]
#             j2 = suffix__array[i - 1]
#             if i:
#                 while l + j < endAt[j] and l + j2 < endAt[j2] and s[j + l] == s[j2 + l]:
#                     l += 1
#                 LCP[i - 1] = l
#             else:
#                 l = 0
#         self.lcp = LCP

#     def _step3_rstr(self):
#         prev_len = 0
#         idx = 0
#         results = {}
#         len_lcp = len(self.lcp) - 1

#         class Stack:
#             pass

#         stack = Stack()
#         stack._top = 0
#         stack.lst_max = []

#         # if len(self.res) == 0 :
#         #  return {}

#         pos1 = self.res[0]
#         for idx in range(len_lcp):
#             current_len = self.lcp[idx]
#             pos2 = self.res[idx + 1]
#             end_ = max(pos1, pos2) + current_len
#             n = prev_len - current_len
#             if n < 0:
#                 # pushMany
#                 stack.lst_max.append([-n, idx, end_])
#                 stack._top += -n
#             elif n > 0:
#                 self.removeMany(stack, results, n, idx)
#             elif stack._top > 0 and end_ > stack.lst_max[-1][-1]:
#                 # setMax
#                 stack.lst_max[-1][-1] = end_

#             prev_len = current_len
#             pos1 = pos2

#         if stack._top > 0:
#             self.removeMany(stack, results, stack._top, idx + 1)

#         return results

#     def removeMany(self, stack, results, m, idxEnd):
#         prevStart = -1
#         while m > 0:
#             n, idxStart, maxEnd = stack.lst_max.pop()
#             if prevStart != idxStart:
#                 # idStr = self.idxString[maxEnd-1]
#                 # pos = self.idxPos[maxEnd-1]
#                 id_ = (maxEnd, idxEnd - idxStart + 1)
#                 if id_ not in results or results[id_][0] < stack._top:
#                     results[id_] = (stack._top, idxStart)
#                 prevStart = idxStart
#             m -= n
#             stack._top -= n
#         if m < 0:
#             stack.lst_max.append([-m, idxStart, maxEnd - n - m])
#             stack._top -= m

#     def go(self):
#         self._step1_sort_suffix()
#         self._step2_lcp()
#         return self._step3_rstr()


# def common_sub_strings_py(stringx: str, stringy: str, limit=25):
#     """Finds all common substrings between stringx and stringy
#     longer than limit. This function is case sensitive.
#     The substrings may overlap.

#     returns a list of tuples describing the substrings
#     The list is sorted longest -> shortest.

#     Parameters
#     ----------
#     stringx : str
#     stringy : str
#     limit : int, optional

#     Returns
#     -------
#     list of tuple
#         [(startx1, starty1, length1),(startx2, starty2, length2), ...]

#         startx1 = startposition in x, where substring 1 starts
#         starty1 = position in y where substring 1 starts
#         length1 = lenght of substring


#     Examples
#     --------

#     >>> from pydna.common_sub_strings import common_sub_strings
#     >>> common_sub_strings("gatgatttcggtagtta", "gtcagtatgtctatctatcgcg", limit=3)
#     [(1, 6, 3), (7, 17, 3), (10, 4, 3), (12, 3, 3)]

#     ::

#         Overlaps   Symbols
#         (1, 6,  3)   ---
#         (7, 17, 3)   +++
#         (10, 4, 3)   ...
#         (12, 3, 3)   ===


#                     ===
#         gatgatttcggtagtta           stringx
#          ---   +++...

#             ...
#         gtcagtatgtctatctatcgcg      stringy
#            ===---        +++

#     """
#     rstr = _Rstr_max()
#     rstr.add_str("&".join((stringx, stringy)))
#     r = rstr.go()
#     match = {}
#     for (offset_end, nb), (l, start_plage) in r.items():
#         if l < limit:
#             continue
#         startsx = []
#         startsy = []
#         for o in range(start_plage, start_plage + nb):
#             offset = rstr.idxPos[rstr.res[o]]
#             if offset > len(stringx):
#                 startsy.append(offset - len(stringx) - 1)
#             else:
#                 startsx.append(offset)

#         for a, b in _itertools.product(startsx, startsy):
#             match[(a, b)] = max(match.get((a, b)) or 0, l)

#     match = [(key[0], key[1], val) for key, val in list(match.items())]

#     match.sort()

#     match.sort(key=_itemgetter(2), reverse=True)

#     return match


def common_sub_strings(stringx: str, stringy: str, limit: int = 25) -> _List[Match]:
    """
    Finds all common substrings between stringx and stringy, and returns
    them sorted by length.

    This function is case sensitive.

    Parameters
    ----------
    stringx : str
    stringy : str
    limit : int, optional

    Returns
    -------
    list of tuple
        [(startx1, starty1, length1),(startx2, starty2, length2), ...]

        startx1 = startposition in x, where substring 1 starts
        starty1 = position in y where substring 1 starts
        length1 = lenght of substring
    """
    from pydivsufsort import common_substrings

    matches = common_substrings(stringx, stringy, limit=limit)
    matches.sort()
    matches.sort(key=_itemgetter(2), reverse=True)
    return matches


def terminal_overlap(stringx: str, stringy: str, limit: int = 15) -> _List[Match]:
    """Finds the the flanking common substrings between stringx and stringy
    longer than limit. This means that the results only contains substrings
    that starts or ends at the the ends of stringx and stringy.

    This function is case sensitive.

    returns a list of tuples describing the substrings
    The list is sorted longest -> shortest.

    Parameters
    ----------
    stringx : str
    stringy : str
    limit : int, optional

    Returns
    -------
    list of tuple
        [(startx1,starty1,length1),(startx2,starty2,length2), ...]

        startx1 = startposition in x, where substring 1 starts
        starty1 = position in y where substring 1 starts
        length1 = lenght of substring


    Examples
    --------

    >>> from pydna.common_sub_strings import terminal_overlap
    >>> terminal_overlap("agctatgtatcttgcatcgta", "gcatcgtagtctatttgcttac", limit=8)
    [(13, 0, 8)]

    ::

                        <-- 8 ->
           <---- 13 --->
           agctatgtatcttgcatcgta                    stringx
                        gcatcgtagtctatttgcttac      stringy
                        0

    """
    return [
        m
        for m in common_sub_strings(stringx, stringy, limit)
        if (m[0] == 0 and m[1] + m[2] == len(stringy))
        or (m[1] == 0 and m[0] + m[2] == len(stringx))
    ]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from pydna.common_sub_strings import terminal_overlap
from pydna.dseq import Dseq
from pydna.dseqrecord import Dseqrecord
from pydna.utils import rc
from itertools import combinations
from copy import copy


def fuse_by_pcr(fragments, limit=15):
    """docstring."""

    def anneal(x, y, limit=limit):
        """docstring."""
        new = None
        for a, b in [(x, y), (x, y.rc()), (x.rc(), y)]:
            try:
                ((s1, s2, ln), *r) = terminal_overlap(
                    a.seq.watson.lower(), rc(b.seq.crick.lower()), limit=limit
                )
            except ValueError as err:
                if "not enough values to unpack" not in str(err):
                    raise err
            else:
                if s2 == 0:
                    new = Dseqrecord(Dseq(a.seq.watson, b.seq.crick, ovhg=-s1))
                    new.features = copy(a.features)
                    new.features.extend([f._shift(s1) for f in b.features])
                    new.left = b  # Setting a new property dynamically
                    new.right = a  # Setting a new property dynamically
                elif s1 == 0:
                    new = Dseqrecord(Dseq(b.seq.watson, a.seq.crick, ovhg=-s2))
                    new.features = copy(b.features)
                    new.features.extend([f._shift(s2) for f in a.features])
                    new.left = a  # Setting a new property dynamically
                    new.right = b  # Setting a new property dynamically
        return new

    argument = fragments
    for arg in argument:
        arg.left = None
        arg.right = None
    newfragments = []
    while True:
        for a, b in combinations(fragments, 2):
            new = anneal(a, b, limit)
            if not new:
                continue
            new.anneal = new.seq
            new.seq = new.seq.fill_in()
            new.features = list({repr(f): f for f in new.features}.values())
            newfragments.append(new)
        if newfragments:
            fragments = newfragments
            newfragments = []
        else:
            break

    return [x for x in fragments if x not in argument]


def list_parts(fusion_pcr_fragment):
    stack = [fusion_pcr_fragment]
    processed = []

    while stack:
        obj = stack.pop()
        try:
            a, b = obj.right, obj.left
        except AttributeError:
            pass
        else:
            stack.extend((a, b))
        processed.append(obj)

    parts = [x for x in processed[::-1] if x]

    msg = "---\n\n"

    for part in parts:
        if hasattr(part, "anneal"):
            msg += repr(part.anneal) + "\n\n"
            msg += f"{part.name}\n{part.seq.watson}\n{part.seq.crick[::-1]}\n\n---\n\n"
        msg += repr(part.seq) + "\n\n"

    return msg
